/**********************************************************************
project: <Universal Accessory Decoder>
author: <Thierry PARIS>
description: <Class for a light, flashing or not, with optional fading>
***********************************************************************/

#include "UniversalAccessoryDecoder.h"
#include "AccessoryLight.hpp"

#ifndef NO_LIGHT

AccessoryLight::AccessoryLight(int inDccId, byte inDccIdAccessory, unsigned long inBlinkDuration) : Accessory(inDccId, inDccIdAccessory, false, inBlinkDuration)
{
	this->flashingState = false;
	this->fadingStep = this->fadingDelay = 0;
}

void AccessoryLight::SetState(ACC_STATE inState) 
{ 
#ifdef DEBUG_MODE
	Serial.print(F("AccessoryLight SetState "));
	Serial.println(inState == LIGHTON ? "ON" : inState == LIGHTOFF ? "OFF" : "BLINK");
#endif

	if (inState == LIGHTBLINK && this->GetDuration() == 0)
		inState = LIGHTON;

	if (this->state == inState)
	{
		if (inState == LIGHTBLINK)
		{
			// If the state is not changing, but the state is blinking, reset the blink to restart from beginning.
			// If this led is in a group, all blinking lights in this group will start blinking at the same time.
			// TODO : one day, add a flag to decide to not synchronize all blinking leds in a group...
			this->ResetAction();
			this->LightOnRaw();
			this->flashingState = true;
			this->StartAction();
		}
		return;
	}

	switch (inState)
	{
		case LIGHTON:
			if (this->IsFading())
			{
				this->fadingCurrentDirection = 1;
				this->fadingCurrentValue = 0;
				this->StartAction();
				break;
			}

			this->LightOn();
			this->flashingState = false;
			this->ResetAction();
			break;

		case LIGHTBLINK:
			this->LightOnRaw();
			this->flashingState = true;
			this->StartAction();
			break;

		default:
			if (this->IsFading())
			{
				this->fadingCurrentDirection = -1;
				this->fadingCurrentValue = this->pDriver->GetSpeed(MOTOR_LIGHT, this->port);
				this->StartAction();
				break;
			}
			this->LightOff();
			this->ResetAction();
	}

	this->state = inState;
}

void AccessoryLight::SetFading(unsigned char inStep, unsigned char inDelay)
{
	this->fadingStep = inStep;
	this->fadingDelay = inDelay;
	this->fadingCurrentDirection = 0;
	this->fadingCurrentValue = 0;
}

void AccessoryLight::Setup(Driver *inpDriver, unsigned char inPort, int inIntensity)
{
	this->pDriver = inpDriver;
	this->port = inPort;
	this->pDriver->SetSpeed(MOTOR_LIGHT, this->port, inIntensity);
	this->LightOff();
}

void AccessoryLight::LightFading(unsigned char inValue)
{
#ifdef DEBUG_MODE
#ifdef DEBUG_VERBOSE_MODE
	Serial.print(F("AccessoryLight Fading at "));
	Serial.println(inValue);
#endif
#endif
	this->pDriver->MoveLeftDir(MOTOR_LIGHT, this->port, -1, inValue);
}

void AccessoryLight::LightOnRaw()
{
	this->pDriver->MoveLeftDir(MOTOR_LIGHT, this->port, -1, this->pDriver->GetSpeed(MOTOR_LIGHT, this->port));
}

void AccessoryLight::LightOffRaw()
{
	this->pDriver->MoveStop(MOTOR_LIGHT, this->port);
}

void AccessoryLight::LightOn()
{
#ifdef DEBUG_MODE
	Serial.println(F("AccessoryLight ON"));
#endif
	this->LightOnRaw();
	this->state = LIGHTON;
}

void AccessoryLight::LightOff()
{
#ifdef DEBUG_MODE
	Serial.println(F("AccessoryLight OFF"));
#endif
	this->LightOffRaw();
	this->state = LIGHTOFF;
}

ACC_STATE AccessoryLight::Toggle()
{
	if (this->state != LIGHTBLINK)
	{
		if (this->IsFading())
		{
			// Start a new fading.
			if (this->state == LIGHTON)
			{
				this->fadingCurrentDirection = -1;
				this->fadingCurrentValue = this->pDriver->GetSpeed(MOTOR_LIGHT, this->port);
				this->StartAction();
				return this->state;
			}
			if (this->state == LIGHTOFF)
			{
				this->fadingCurrentDirection = 1;
				this->fadingCurrentValue = 0;
				this->StartAction();
				return this->state;
			}
		}
	}

	if (this->state == LIGHTON || this->state == LIGHTBLINK)
		this->SetState(LIGHTOFF);
	else
		this->SetState(LIGHTBLINK);	// will be converted in LIGHTON if Duration is 0 !

	return this->state; 
}

void AccessoryLight::Move(int inDccId, byte inDccAccessory)
{
	if (this->GetDccPositionSize() == 1)
	{
		this->Toggle();
		return;
	}

	int position = this->GetDccPosition(inDccId, inDccAccessory);

	if (position != -1)
		this->SetState((ACC_STATE) position);
}

void AccessoryLight::StartAction()
{
	if (this->GetDuration() > 0 || this->IsFading())
		this->SetStartingMillis();

#ifdef DEBUG_MODE
#ifdef DEBUG_VERBOSE_MODE
	Serial.print(F("AccessoryLight start action "));
	Serial.println(this->GetActionStartingMillis());
#endif
#endif
}

bool AccessoryLight::ActionEnded()
{
#ifdef DEBUG_MODE
#ifdef DEBUG_VERBOSE_MODE
	Serial.print(F("End action of "));
	Serial.println(F("light."));
#endif
#endif
	if (this->IsFading())
	{
		if (this->GetActionStartingMillis() > 0)
		{
			if (millis() - this->GetActionStartingMillis() > this->fadingDelay)
			{
				int curr = (int) this->fadingCurrentValue + this->fadingStep * this->fadingCurrentDirection;
				if (curr > this->pDriver->GetSpeed(MOTOR_LIGHT, this->port) || curr < 0)
				{
					// End of fading
					this->fadingCurrentDirection = 0;
					this->ResetAction();
					if (curr > this->pDriver->GetSpeed(MOTOR_LIGHT, this->port))
						this->LightOn();
					if (curr < 0)
						this->LightOff();
					return true;
				}
				this->fadingCurrentValue = (unsigned char)curr;
				this->LightFading(curr);
				this->StartAction();
				return false;
			}
		}
	}

	if (this->state == LIGHTBLINK)
	{
		if (this->GetActionStartingMillis() > 0)
		{
			if (millis() - this->GetActionStartingMillis() > this->GetDuration())
			{
#ifdef DEBUG_MODE
#ifdef DEBUG_VERBOSE_MODE
				Serial.print(F("End action of "));
				Serial.print(F("light at "));
				Serial.print(millis() - this->GetActionStartingMillis());
				Serial.print(F("ms for "));
				Serial.print(this->GetDuration());
				Serial.println(F("ms"));
#endif
#endif
				if (flashingState)
					this->LightOffRaw();
				else
					this->LightOnRaw();
				flashingState = !flashingState;
				this->StartAction();
			}
		}
	}

	return false;
}
#endif
